/** 
 * DevExtreme (ui/scheduler/ui.scheduler.resource_manager.js)
 * Version: 16.1.4
 * Build date: Wed Jun 22 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var $ = require("jquery"),
        Class = require("../../core/class"),
        arrayUtils = require("../../core/utils/array"),
        commonUtils = require("../../core/utils/common"),
        query = require("../../data/query"),
        dataCoreUtils = require("../../core/utils/data"),
        DataSourceModule = require("../../data/data_source/data_source");
    var DEFAULT_VALUE_EXPR = "id",
        DEFAULT_DISPLAY_EXPR = "text";
    var ResourceManager = Class.inherit({
        _wrapDataSource: function(dataSource) {
            if (dataSource instanceof DataSourceModule.DataSource) {
                return dataSource
            } else {
                return new DataSourceModule.DataSource({
                    store: DataSourceModule.normalizeDataSourceOptions(dataSource).store,
                    pageSize: 0
                })
            }
        },
        _mapResourceData: function(resource, data) {
            var valueGetter = dataCoreUtils.compileGetter(resource.valueExpr || DEFAULT_VALUE_EXPR),
                displayGetter = dataCoreUtils.compileGetter(resource.displayExpr || DEFAULT_DISPLAY_EXPR);
            return $.map(data, function(item) {
                var result = {
                    id: valueGetter(item),
                    text: displayGetter(item)
                };
                if (item.color) {
                    result.color = item.color
                }
                return result
            })
        },
        _isMultipleResource: function(resourceField) {
            var result = false;
            $.each(this.getResources(), $.proxy(function(_, resource) {
                var field = this.getField(resource);
                if (field === resourceField) {
                    result = resource.allowMultiple;
                    return false
                }
            }, this));
            return result
        },
        ctor: function(resources) {
            this.setResources(resources)
        },
        setResources: function(resources) {
            this._resources = resources;
            this._dataAccessors = {
                getter: {},
                setter: {}
            };
            this._resourceFields = $.map(resources || [], $.proxy(function(resource, index) {
                var field = this.getField(resource);
                this._dataAccessors.getter[field] = dataCoreUtils.compileGetter(field);
                this._dataAccessors.setter[field] = dataCoreUtils.compileSetter(field);
                return field
            }, this))
        },
        getResources: function() {
            return this._resources || []
        },
        getEditors: function(loadedResources) {
            var result = [],
                that = this;
            loadedResources = loadedResources || [];
            $.each(this.getResources(), function(i, resource) {
                var field = that.getField(resource),
                    currentResourceItems = that._getLoadedResourceItemsByField(loadedResources, field);
                result.push({
                    editorOptions: {
                        dataSource: currentResourceItems.length ? currentResourceItems : that._wrapDataSource(resource.dataSource),
                        displayExpr: resource.displayExpr || DEFAULT_DISPLAY_EXPR,
                        valueExpr: resource.valueExpr || DEFAULT_VALUE_EXPR
                    },
                    dataField: field,
                    editorType: resource.allowMultiple ? "dxTagBox" : "dxSelectBox",
                    label: {
                        text: resource.label || field
                    }
                })
            });
            return result
        },
        getDataAccessors: function(field, type) {
            var result = null;
            $.each(this._dataAccessors[type], function(accessorName, accessors) {
                if (field === accessorName) {
                    result = accessors;
                    return false
                }
            });
            return result
        },
        getResourceDataByValue: function(field, value) {
            var result = $.Deferred(),
                that = this;
            $.each(this.getResources(), function(_, resource) {
                var resourceField = that.getField(resource);
                if (resourceField === field) {
                    var dataSource = that._wrapDataSource(resource.dataSource),
                        valueExpr = resource.valueExpr || DEFAULT_VALUE_EXPR;
                    dataSource.load().done(function(data) {
                        var filteredData = query(data).filter(valueExpr, value).toArray();
                        result.resolve(filteredData[0])
                    }).fail(function() {
                        result.reject()
                    });
                    return false
                }
            });
            return result.promise()
        },
        getResourcesFromItem: function(itemData, wrapOnlyMultipleResources) {
            var that = this,
                result = null;
            if (!commonUtils.isDefined(wrapOnlyMultipleResources)) {
                wrapOnlyMultipleResources = false
            }
            $.each(that._resourceFields, function(index, field) {
                $.each(itemData, function(fieldName, fieldValue) {
                    var tmp = {};
                    tmp[fieldName] = fieldValue;
                    var resourceData = that.getDataAccessors(field, "getter")(tmp);
                    if (resourceData) {
                        if (!result) {
                            result = {}
                        }
                        if (!wrapOnlyMultipleResources || wrapOnlyMultipleResources && that._isMultipleResource(field)) {
                            that.getDataAccessors(field, "setter")(tmp, arrayUtils.wrapToArray(resourceData))
                        } else {
                            that.getDataAccessors(field, "setter")(tmp, resourceData)
                        }
                        $.extend(result, tmp);
                        return true
                    }
                })
            });
            return result
        },
        getResourcesValueByFields: function(groups) {
            var result = $.Deferred(),
                that = this,
                deferreds = [];
            $.each(this.getResourcesByFields(groups), function(i, resource) {
                var deferred = $.Deferred(),
                    field = that.getField(resource);
                deferreds.push(deferred);
                that._wrapDataSource(resource.dataSource).load().done(function(data) {
                    deferred.resolve({
                        name: field,
                        items: that._mapResourceData(resource, data)
                    })
                }).fail(function() {
                    deferred.reject()
                })
            });
            $.when.apply(null, deferreds).done(function() {
                result.resolve(Array.prototype.slice.call(arguments))
            }).fail(function() {
                result.reject()
            });
            return result.promise()
        },
        getResourceColor: function(field, value) {
            var color, result = $.Deferred();
            this.getResourceDataByValue(field, value).done(function(data) {
                if (data) {
                    color = data.color
                }
                result.resolve(color)
            }).fail(function() {
                result.reject()
            });
            return result.promise()
        },
        getResourcesByFields: function(fields) {
            return $.grep(this.getResources(), $.proxy(function(resource) {
                var field = this.getField(resource);
                return $.inArray(field, fields) > -1
            }, this))
        },
        getResourceForPainting: function(groups) {
            var result, resources = this.getResources();
            $.each(resources, function(index, resource) {
                if (resource.useColorAsDefault || resource.mainColor) {
                    result = resource;
                    return false
                }
            });
            if (!result) {
                if ($.isArray(groups) && groups.length) {
                    resources = this.getResourcesByFields(groups)
                }
                result = resources[resources.length - 1]
            }
            return result
        },
        createResourcesTree: function(groups) {
            var leafIndex = 0,
                groupIndex = groupIndex || 0;

            function make(group, groupIndex, result, parent) {
                result = result || [];
                for (var i = 0; i < group.items.length; i++) {
                    var currentGroupItem = group.items[i];
                    var resultItem = {
                        name: group.name,
                        value: currentGroupItem.id,
                        title: currentGroupItem.text,
                        children: [],
                        parent: parent ? parent : null
                    };
                    result.push(resultItem);
                    var nextGroupIndex = groupIndex + 1;
                    if (groups[nextGroupIndex]) {
                        make.call(this, groups[nextGroupIndex], nextGroupIndex, resultItem.children, resultItem)
                    }
                    if (!resultItem.children.length) {
                        resultItem.leafIndex = leafIndex;
                        leafIndex++
                    }
                }
                return result
            }
            return make.call(this, groups[0], 0)
        },
        _hasGroupItem: function(appointmentResources, groupName, itemValue) {
            var group = this.getDataAccessors(groupName, "getter")(appointmentResources);
            if (group) {
                if ($.inArray(itemValue, group) > -1) {
                    return true
                }
            }
            return false
        },
        _getLoadedResourceItemsByField: function(loadedResources, fieldName) {
            var currentResourceItems = [];
            for (var i = 0, resourceCount = loadedResources.length; i < resourceCount; i++) {
                if (loadedResources[i].name === fieldName) {
                    currentResourceItems = loadedResources[i].items;
                    break
                }
            }
            return currentResourceItems
        },
        getResourceTreeLeaves: function(tree, appointmentResources, result) {
            result = result || [];
            for (var i = 0; i < tree.length; i++) {
                if (!this._hasGroupItem(appointmentResources, tree[i].name, tree[i].value)) {
                    continue
                }
                if (commonUtils.isDefined(tree[i].leafIndex)) {
                    result.push(tree[i].leafIndex)
                }
                if (tree[i].children) {
                    this.getResourceTreeLeaves(tree[i].children, appointmentResources, result)
                }
            }
            return result
        },
        groupAppointmentsByResources: function(appointments, resources) {
            var tree = this.createResourcesTree(resources),
                result = {};
            $.each(appointments, $.proxy(function(_, appointment) {
                var appointmentResources = this.getResourcesFromItem(appointment),
                    treeLeaves = this.getResourceTreeLeaves(tree, appointmentResources);
                for (var i = 0; i < treeLeaves.length; i++) {
                    if (!result[treeLeaves[i]]) {
                        result[treeLeaves[i]] = []
                    }
                    result[treeLeaves[i]].push(appointment)
                }
            }, this));
            return result
        },
        reduceResourcesTree: function(tree, existingAppointments, _result) {
            _result = _result ? _result.children : [];
            var that = this;
            tree.forEach(function(node, index) {
                var ok = false,
                    resourceName = node.name,
                    resourceValue = node.value,
                    resourceTitle = node.title,
                    resourceGetter = that.getDataAccessors(resourceName, "getter");
                existingAppointments.forEach(function(appointment) {
                    if (!ok) {
                        var resourceFromAppointment = resourceGetter(appointment);
                        if (commonUtils.isArray(resourceFromAppointment)) {
                            if (resourceFromAppointment.indexOf(resourceValue) > -1) {
                                _result.push({
                                    name: resourceName,
                                    value: resourceValue,
                                    title: resourceTitle,
                                    children: []
                                });
                                ok = true
                            }
                        } else {
                            if (resourceFromAppointment === resourceValue) {
                                _result.push({
                                    name: resourceName,
                                    value: resourceValue,
                                    title: resourceTitle,
                                    children: []
                                });
                                ok = true
                            }
                        }
                    }
                });
                if (ok && node.children && node.children.length) {
                    that.reduceResourcesTree(node.children, existingAppointments, _result[index])
                }
            });
            return _result
        },
        getField: function(resource) {
            return resource.fieldExpr || resource.field
        },
        getColorFromLoadedResources: function(loadedResources, fieldName, value) {
            var result = $.Deferred(),
                requiredResourceLoaded = false,
                currentResourceItems = this._getLoadedResourceItemsByField(loadedResources, fieldName);
            if (currentResourceItems.length) {
                requiredResourceLoaded = true
            }
            if (requiredResourceLoaded) {
                for (var j = 0, itemCount = currentResourceItems.length; j < itemCount; j++) {
                    if (currentResourceItems[j].id === value) {
                        result.resolve(currentResourceItems[j].color);
                        break
                    }
                }
            } else {
                this.getResourceDataByValue(fieldName, value).done(function(resourceData) {
                    result.resolve(resourceData.color)
                })
            }
            return result.promise()
        }
    });
    module.exports = ResourceManager
});
