/** 
 * DevExtreme (viz/components/data_validator.js)
 * Version: 16.1.4
 * Build date: Wed Jun 22 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var $ = require("jquery"),
        commonUtils = require("../../core/utils/common"),
        STRING = "string",
        NUMERIC = "numeric",
        DATETIME = "datetime",
        DISCRETE = "discrete",
        SEMIDISCRETE = "semidiscrete",
        CONTINUOUS = "continuous",
        LOGARITHMIC = "logarithmic",
        VALUE_TYPE = "valueType",
        ARGUMENT_TYPE = "argumentType",
        axisTypeParser = require("../core/utils").enumParser([STRING, NUMERIC, DATETIME]),
        _getParser = require("./parse_utils").getParser,
        _isDefined = commonUtils.isDefined,
        _isFunction = commonUtils.isFunction,
        _isArray = commonUtils.isArray,
        _isString = commonUtils.isString,
        _isDate = commonUtils.isDate,
        _isNumber = commonUtils.isNumber,
        _isObject = commonUtils.isObject,
        _each = $.each;

    function groupingValues(data, others, valueField, index) {
        if (index >= 0) {
            _each(data.slice(index), function(_, cell) {
                if (_isDefined(cell[valueField])) {
                    others[valueField] += cell[valueField];
                    cell[valueField] = cell["original" + valueField] = void 0
                }
            })
        }
    }

    function mergeSort(data, field) {
        function merge_sort(array, low, high, field) {
            if (low < high) {
                var mid = Math.floor((low + high) / 2);
                merge_sort(array, low, mid, field);
                merge_sort(array, mid + 1, high, field);
                merge(array, low, mid, high, field)
            }
        }
        var n = data.length;
        merge_sort(data, 0, n - 1, field);
        return data
    }

    function merge(array, low, mid, high, field) {
        var k, newArray = new Array(high - low + 1),
            countL = low,
            countR = mid + 1,
            i = 0;
        while (countL <= mid && countR <= high) {
            if (array[countL][field] <= array[countR][field] || !_isDefined(array[countR][field])) {
                newArray[i] = array[countL];
                countL++
            } else {
                newArray[i] = array[countR];
                countR++
            }
            i++
        }
        if (countL > mid) {
            for (k = countR; k <= high; k++, i++) {
                newArray[i] = array[k]
            }
        } else {
            for (k = countL; k <= mid; k++, i++) {
                newArray[i] = array[k]
            }
        }
        for (k = 0; k <= high - low; k++) {
            array[k + low] = newArray[k]
        }
        return array
    }

    function processGroup(_, group) {
        group.valueType = group.valueAxisType = null;
        _each(group, processSeries);
        group.valueAxis && group.valueAxis.resetTypes(VALUE_TYPE)
    }

    function parseCategories(categories, parser) {
        var newArray = [];
        _each(categories, function(_, category) {
            var parsedCategory = parser(category);
            void 0 !== parsedCategory && newArray.push(parsedCategory)
        });
        return newArray
    }

    function parseAxisCategories(groups, parsers) {
        var argumentCategories = groups.argumentOptions && groups.argumentOptions.categories,
            valueParser = parsers[1];
        _each(groups, function(_, valueGroup) {
            var categories = valueGroup.valueOptions && valueGroup.valueOptions.categories;
            if (categories) {
                valueGroup.valueOptions.categories = parseCategories(categories, valueParser)
            }
        });
        if (argumentCategories) {
            groups.argumentOptions.categories = parseCategories(argumentCategories, parsers[0])
        }
    }

    function processSeries(_, series) {
        series.updateDataType({})
    }

    function resetAxisTypes(_, axis) {
        axis.resetTypes(ARGUMENT_TYPE)
    }

    function filterForLogAxis(val, field, incidentOccurred) {
        if (val <= 0) {
            incidentOccurred("E2004", [field]);
            val = null
        }
        return val
    }

    function eigen(x) {
        return x
    }

    function getType(unit, type) {
        var result = type;
        if (type === STRING || _isString(unit)) {
            result = STRING
        } else {
            if (type === DATETIME || _isDate(unit)) {
                result = DATETIME
            } else {
                if (_isNumber(unit)) {
                    result = NUMERIC
                }
            }
        }
        return result
    }

    function correctAxisType(type, axisType, hasCategories, incidentOccurred) {
        if (type === STRING && (axisType === CONTINUOUS || axisType === LOGARITHMIC || axisType === SEMIDISCRETE)) {
            incidentOccurred("E2002")
        }
        return axisType === LOGARITHMIC ? LOGARITHMIC : hasCategories || axisType === DISCRETE || type === STRING ? DISCRETE : axisType === SEMIDISCRETE ? SEMIDISCRETE : CONTINUOUS
    }

    function validUnit(unit, field, incidentOccurred) {
        if (unit) {
            incidentOccurred(!_isNumber(unit) && !_isDate(unit) && !_isString(unit) ? "E2003" : "E2004", [field])
        }
    }

    function createParserUnit(type, axisType, ignoreEmptyPoints, skipFields, incidentOccurred) {
        var parser = type ? _getParser(type) : eigen,
            filter = axisType === LOGARITHMIC ? filterForLogAxis : eigen;
        return function(unit, field) {
            var parseUnit = filter(parser(unit), field, incidentOccurred);
            null === parseUnit && ignoreEmptyPoints && (parseUnit = void 0);
            if (void 0 === parseUnit) {
                skipFields[field] = (skipFields[field] || 0) + 1;
                validUnit(unit, field, incidentOccurred)
            }
            return parseUnit
        }
    }

    function prepareParsers(groups, skipFields, incidentOccurred) {
        var sizeParser, valueParser, iep, argumentParser = createParserUnit(groups.argumentType, groups.argumentAxisType, false, skipFields, incidentOccurred),
            categoryParsers = [argumentParser],
            cache = {},
            list = [];
        _each(groups, function(_, group) {
            _each(group, function(_, series) {
                iep = series.getOptions().ignoreEmptyPoints;
                valueParser = createParserUnit(group.valueType, group.valueAxisType, iep, skipFields, incidentOccurred);
                sizeParser = createParserUnit(NUMERIC, CONTINUOUS, iep, skipFields, incidentOccurred);
                cache[series.getArgumentField()] = argumentParser;
                _each(series.getValueFields(), function(_, field) {
                    !categoryParsers[1] && (categoryParsers[1] = valueParser);
                    cache[field] = valueParser
                });
                if (series.getSizeField()) {
                    cache[series.getSizeField()] = sizeParser
                }
                if (series.getTagField()) {
                    cache[series.getTagField()] = eigen
                }
            })
        });
        _each(cache, function(field, parser) {
            list.push([field, parser])
        });
        list.length && parseAxisCategories(groups, categoryParsers);
        return list
    }

    function getParsedCell(cell, parsers) {
        var i, field, value, ii = parsers.length,
            obj = {};
        for (i = 0; i < ii; ++i) {
            field = parsers[i][0];
            value = cell[field];
            obj[field] = parsers[i][1](value, field);
            obj["original" + field] = value
        }
        return obj
    }

    function parse(data, parsers) {
        var i, parsedData = [],
            ii = data.length;
        parsedData.length = ii;
        for (i = 0; i < ii; ++i) {
            parsedData[i] = getParsedCell(data[i], parsers)
        }
        return parsedData
    }

    function findIndexByThreshold(data, valueField, threshold) {
        var i, value, ii = data.length;
        for (i = 0; i < ii; ++i) {
            value = data[i][valueField];
            if (_isDefined(value) && threshold > value) {
                break
            }
        }
        return i
    }

    function groupMinSlices(originalData, argumentField, valueField, smallValuesGrouping) {
        smallValuesGrouping = smallValuesGrouping || {};
        var data, mode = smallValuesGrouping.mode,
            others = {};
        if (!mode || "none" === mode) {
            return
        }
        others[argumentField] = String(smallValuesGrouping.groupName || "others");
        others[valueField] = 0;
        data = originalData.slice();
        data.sort(function(a, b) {
            var isA = _isDefined(a[valueField]) ? 1 : 0,
                isB = _isDefined(b[valueField]) ? 1 : 0;
            return isA && isB ? b[valueField] - a[valueField] : isB - isA
        });
        groupingValues(data, others, valueField, "smallValueThreshold" === mode ? findIndexByThreshold(data, valueField, smallValuesGrouping.threshold) : smallValuesGrouping.topCount);
        others[valueField] && originalData.push(others)
    }

    function groupData(data, groups) {
        var isPie = groups[0] && groups[0][0] && ("pie" === groups[0][0].type || "doughnut" === groups[0][0].type || "donut" === groups[0][0].type);
        if (!isPie) {
            return
        }
        _each(groups, function(_, group) {
            _each(group, function(_, series) {
                groupMinSlices(data, series.getArgumentField(), series.getValueFields()[0], series.getOptions().smallValuesGrouping)
            })
        })
    }

    function sort(data, groups, sortingMethod, userArgumentCategories) {
        var hash = {},
            argumentField = groups.length && groups[0].length && groups[0][0].getArgumentField();
        if (_isFunction(sortingMethod)) {
            data.sort(sortingMethod)
        } else {
            if (userArgumentCategories.length) {
                _each(userArgumentCategories, function(index, value) {
                    hash[value] = index
                });
                data.sort(function(a, b) {
                    a = a[argumentField];
                    b = b[argumentField];
                    return hash[a] - hash[b]
                })
            } else {
                if (true === sortingMethod && groups.argumentType !== STRING) {
                    mergeSort(data, argumentField)
                }
            }
        }
    }

    function checkValueTypeOfGroup(group, cell) {
        _each(group, function(_, series) {
            _each(series.getValueFields(), function(_, field) {
                group.valueType = getType(cell[field], group.valueType)
            })
        });
        return group.valueType
    }

    function checkArgumentTypeOfGroup(group, cell, groups) {
        _each(group, function(_, series) {
            groups.argumentType = getType(cell[series.getArgumentField()], groups.argumentType)
        });
        return groups.argumentType
    }

    function checkType(data, groups, checkTypeForAllData) {
        var groupsWithUndefinedValueType = [],
            groupsWithUndefinedArgumentType = [],
            argumentTypeGroup = groups.argumentOptions && axisTypeParser(groups.argumentOptions.argumentType);
        _each(groups, function(_, group) {
            if (!group.length) {
                return null
            }
            var valueTypeGroup = group.valueOptions && axisTypeParser(group.valueOptions.valueType);
            group.valueType = valueTypeGroup;
            groups.argumentType = argumentTypeGroup;
            !valueTypeGroup && groupsWithUndefinedValueType.push(group);
            !argumentTypeGroup && groupsWithUndefinedArgumentType.push(group)
        });
        if (groupsWithUndefinedValueType.length || groupsWithUndefinedArgumentType.length) {
            _each(data, function(_, cell) {
                var defineVal, defineArg;
                _each(groupsWithUndefinedValueType, function(_, group) {
                    defineVal = checkValueTypeOfGroup(group, cell)
                });
                _each(groupsWithUndefinedArgumentType, function(_, group) {
                    defineArg = checkArgumentTypeOfGroup(group, cell, groups)
                });
                if (!checkTypeForAllData && defineVal && defineArg) {
                    return false
                }
            })
        }
    }

    function checkAxisType(groups, userArgumentCategories, incidentOccurred) {
        var argumentOptions = groups.argumentOptions || {},
            argumentAxisType = correctAxisType(groups.argumentType, argumentOptions.type, !!userArgumentCategories.length, incidentOccurred);
        _each(groups, function(_, group) {
            var valueOptions = group.valueOptions || {},
                valueCategories = valueOptions.categories || [],
                valueAxisType = correctAxisType(group.valueType, valueOptions.type, !!valueCategories.length, incidentOccurred);
            _each(group, function(_, series) {
                var optionsSeries = {};
                optionsSeries.argumentAxisType = argumentAxisType;
                optionsSeries.valueAxisType = valueAxisType;
                groups.argumentAxisType = groups.argumentAxisType || optionsSeries.argumentAxisType;
                group.valueAxisType = group.valueAxisType || optionsSeries.valueAxisType;
                optionsSeries.argumentType = groups.argumentType;
                optionsSeries.valueType = group.valueType;
                optionsSeries.showZero = valueOptions.showZero;
                series.updateDataType(optionsSeries)
            });
            group.valueAxisType = group.valueAxisType || valueAxisType;
            if (group.valueAxis) {
                group.valueAxis.setTypes(group.valueAxisType, group.valueType, VALUE_TYPE);
                group.valueAxis.validate(false)
            }
        });
        groups.argumentAxisType = groups.argumentAxisType || argumentAxisType;
        if (groups.argumentAxes) {
            _each(groups.argumentAxes, function(_, axis) {
                axis.setTypes(groups.argumentAxisType, groups.argumentType, ARGUMENT_TYPE);
                axis.validate(true)
            })
        }
    }

    function verifyData(source, incidentOccurred) {
        var i, ii, k, item, data = [],
            hasError = !_isArray(source);
        if (!hasError) {
            for (i = 0, ii = source.length, k = 0; i < ii; ++i) {
                item = source[i];
                if (_isObject(item)) {
                    data[k++] = item
                } else {
                    if (item) {
                        hasError = true
                    }
                }
            }
        }
        if (hasError) {
            incidentOccurred("E2001")
        }
        return data
    }

    function validateData(data, groups, incidentOccurred, options) {
        var parsers, dataLength, skipFields = {},
            userArgumentCategories = groups.argumentOptions && groups.argumentOptions.categories || [];
        data = verifyData(data, incidentOccurred);
        groups.argumentType = groups.argumentAxisType = null;
        _each(groups, processGroup);
        if (groups.argumentAxes) {
            _each(groups.argumentAxes, resetAxisTypes)
        }
        checkType(data, groups, options.checkTypeForAllData);
        checkAxisType(groups, userArgumentCategories, incidentOccurred);
        if (options.convertToAxisDataType) {
            parsers = prepareParsers(groups, skipFields, incidentOccurred);
            data = parse(data, parsers)
        }
        groupData(data, groups);
        sort(data, groups, options.sortingMethod, userArgumentCategories);
        dataLength = data.length;
        _each(skipFields, function(field, fieldValue) {
            if (fieldValue === dataLength) {
                incidentOccurred("W2002", [field])
            }
        });
        return data
    }
    exports.validateData = validateData
});
