/** 
 * DevExtreme (viz/core/base_widget.js)
 * Version: 16.1.4
 * Build date: Wed Jun 22 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var $ = require("jquery"),
        version = require("../../core/version"),
        _windowResizeCallbacks = require("../../core/utils/window").resizeCallbacks,
        _stringFormat = require("../../core/utils/string").format,
        _Number = Number,
        DOMComponent = require("../../core/dom_component"),
        commonUtils = require("../../core/utils/common"),
        loadingIndicatorModule = require("./loading_indicator"),
        tooltipModule = require("./tooltip"),
        titleModule = require("./title"),
        dataSourceModule = require("./data_source"),
        _isString = commonUtils.isString,
        _parseScalar = require("./utils").parseScalar,
        errors = require("./errors_warnings"),
        _log = errors.log,
        rendererModule = require("./renderers/renderer"),
        headerBlockModule = require("./header_block"),
        exportMenuModule = require("./export"),
        clientExporter = require("../../client_exporter"),
        OPTION_RTL_ENABLED = "rtlEnabled",
        OPTION_LOADING_INDICATOR = "loadingIndicator";

    function getTrue() {
        return true
    }

    function getFalse() {
        return false
    }

    function areCanvasesDifferent(canvas1, canvas2) {
        return !(canvas1.width === canvas2.width && canvas1.height === canvas2.height && canvas1.left === canvas2.left && canvas1.top === canvas2.top && canvas1.right === canvas2.right && canvas1.bottom === canvas2.bottom)
    }

    function createResizeHandler(callback) {
        var timeout, handler = function() {
            clearTimeout(timeout);
            timeout = setTimeout(callback, 100)
        };
        handler.dispose = function() {
            clearTimeout(timeout);
            return this
        };
        return handler
    }

    function defaultOnIncidentOccurred(e) {
        _log.apply(null, [e.target.id].concat(e.target.args || []))
    }
    var createIncidentOccurred = function(widgetName, eventTrigger) {
        return incidentOccurred;

        function incidentOccurred(id, args) {
            eventTrigger("incidentOccurred", {
                target: {
                    id: id,
                    type: "E" === id[0] ? "error" : "warning",
                    args: args,
                    text: _stringFormat.apply(null, [errors.ERROR_MESSAGES[id]].concat(args || [])),
                    widget: widgetName,
                    version: version
                }
            })
        }
    };

    function processTitleOptions(options) {
        var newOptions = _isString(options) ? {
            text: options
        } : options || {};
        newOptions.subtitle = _isString(newOptions.subtitle) ? {
            text: newOptions.subtitle
        } : newOptions.subtitle || {};
        return newOptions
    }

    function pickPositiveValue(value, defaultValue) {
        return _Number(value > 0 ? value : defaultValue || 0)
    }
    module.exports = DOMComponent.inherit({
        _eventsMap: {
            onIncidentOccurred: {
                name: "incidentOccurred"
            },
            onDrawn: {
                name: "drawn"
            },
            onTooltipShown: {
                name: "tooltipShown"
            },
            onTooltipHidden: {
                name: "tooltipHidden"
            },
            onLoadingIndicatorReady: {
                name: "loadingIndicatorReady"
            }
        },
        _getDefaultOptions: function() {
            return $.extend(this.callBase(), {
                onIncidentOccurred: defaultOnIncidentOccurred
            })
        },
        _useLinks: true,
        _init: function() {
            var linkTarget, that = this;
            that.callBase.apply(that, arguments);
            that._themeManager = that._createThemeManager();
            that._themeManager.setCallback(function() {
                that._handleThemeOptions()
            });
            that._initRenderer();
            linkTarget = that._useLinks && that._renderer.root;
            linkTarget && linkTarget.enableLinks().virtualLink("core").virtualLink("peripheral");
            that._renderVisibilityChange();
            that._initEventTrigger();
            that._incidentOccurred = createIncidentOccurred(that.NAME, that._eventTrigger);
            that._initTooltip();
            that._initTitle();
            linkTarget && linkTarget.linkAfter("core");
            that._initDataSource();
            that._initCore();
            linkTarget && linkTarget.linkAfter();
            that._initLoadingIndicator();
            that._initExportMenu();
            that._initHeaderBlock();
            that._setThemeAndRtl();
            that._setContentSize();
            that._setupResizeHandler()
        },
        _initRenderer: function() {
            var that = this;
            that._canvas = that._calculateCanvas();
            that._renderer = new rendererModule.Renderer({
                cssClass: that._rootClassPrefix + " " + that._rootClass,
                pathModified: that.option("pathModified"),
                container: that._$element[0]
            });
            that._renderer.resize(that._canvas.width, that._canvas.height)
        },
        _disposeRenderer: function() {
            this._renderer.dispose()
        },
        _initTitle: function() {
            var that = this;
            that._title = new titleModule.Title({
                renderer: that._renderer,
                cssClass: that._rootClassPrefix + "-title",
                incidentOccurred: that._incidentOccurred
            })
        },
        _disposeTitle: function() {
            this._title.dispose()
        },
        _updateLayoutOfDeltaIndicator: $.noop,
        _getLayoutItems: function() {
            var that = this,
                headerElements = [],
                exportOptions = that._getOption("export"),
                titleOptions = that._title.getLayoutOptions() || {},
                items = [];
            that._exportMenu && exportOptions.enabled && headerElements.push(that._exportMenu);
            "bottom" !== titleOptions.verticalAlignment && headerElements.length ? headerElements.push(that._title) : items.push(that._title);
            if (headerElements.length) {
                that._headerBlock.update(headerElements, that._canvas);
                items.push(that._headerBlock)
            }
            that._updateLayoutOfDeltaIndicator(items);
            return items
        },
        _initTooltip: function() {
            var that = this;
            that._tooltip = new tooltipModule.Tooltip({
                cssClass: that._rootClassPrefix + "-tooltip",
                eventTrigger: that._eventTrigger,
                pathModified: that.option("pathModified")
            })
        },
        _disposeTooltip: function() {
            this._tooltip.dispose()
        },
        _initDataSource: function() {
            var that = this;
            that._dataSource = new dataSourceModule.DataSource(function() {
                that._scheduleLoadingIndicatorHiding();
                that._dataSourceChangedHandler()
            })
        },
        _disposeDataSource: function() {
            this._dataSource.dispose()
        },
        _updateDataSource: function() {
            this._dataSource.update(this.option("dataSource"))
        },
        _getAnimationOptions: $.noop,
        render: function() {
            this._updateSize();
            if (!this._$element.is(":visible")) {
                this._hideTooltip()
            }
        },
        _dispose: function() {
            var that = this;
            that.callBase.apply(that, arguments);
            that._removeResizeHandler();
            that._eventTrigger.dispose();
            that._disposeCore();
            that._disposeDataSource();
            that._disposeLoadingIndicator();
            that._disposeExportMenu();
            that._disposeHeaderBlock();
            that._disposeTooltip();
            that._disposeTitle();
            that._disposeRenderer();
            that._themeManager.dispose();
            that._themeManager = that._renderer = that._eventTrigger = that._tooltip = that._title = that._dataSource = null
        },
        _initEventTrigger: function() {
            var that = this;
            that._eventTrigger = createEventTrigger(that._eventsMap, function(name) {
                return that._createActionByOption(name)
            })
        },
        _calculateCanvas: function() {
            var that = this,
                size = that.option("size") || {},
                margin = that.option("margin") || {},
                defaultCanvas = that._getDefaultSize() || {},
                canvas = {
                    width: size.width <= 0 ? 0 : pickPositiveValue(size.width, that._$element.width() || defaultCanvas.width),
                    height: size.height <= 0 ? 0 : pickPositiveValue(size.height, that._$element.height() || defaultCanvas.height),
                    left: pickPositiveValue(margin.left, defaultCanvas.left || 0),
                    top: pickPositiveValue(margin.top, defaultCanvas.top || 0),
                    right: pickPositiveValue(margin.right, defaultCanvas.right || 0),
                    bottom: pickPositiveValue(margin.bottom, defaultCanvas.bottom || 0)
                };
            if (canvas.width - canvas.left - canvas.right <= 0 || canvas.height - canvas.top - canvas.bottom <= 0) {
                canvas = {
                    width: 0,
                    height: 0
                }
            }
            return canvas
        },
        _updateSize: function() {
            var that = this,
                canvas = that._calculateCanvas();
            if (areCanvasesDifferent(that._canvas, canvas) || that.__forceRender) {
                that._canvas = canvas;
                that._renderer.resize(canvas.width, canvas.height);
                that._setContentSize()
            }
        },
        _setContentSize: function() {
            var nextRect, layoutItems, that = this,
                cache = [],
                canvas = that._canvas,
                rect = canvas.width > 0 && canvas.height > 0 ? [canvas.left, canvas.top, canvas.width - canvas.right, canvas.height - canvas.bottom] : [0, 0, 0, 0];
            that._renderer.lock();
            layoutItems = that._getLayoutItems();
            $.each(layoutItems || [], function(_, layoutItem) {
                var layout = layoutItem.measure([rect[2] - rect[0], rect[3] - rect[1]]);
                if (layout && layout.size[layout.side] < rect[layout.side + 2] - rect[layout.side]) {
                    truncateLayoutArea(rect, layout);
                    cache.push({
                        item: layoutItem,
                        layout: layout
                    })
                }
            });
            nextRect = that._applySize(rect) || rect;
            $.each(cache.reverse(), function(_, cacheItem) {
                var r = buildupLayoutArea(nextRect, cacheItem.layout);
                cacheItem.item.shift(Math.round(r[0]), Math.round(r[1]))
            });
            that._updateLoadingIndicatorSize();
            that._renderer.unlock()
        },
        _getOption: function(name, isScalar) {
            var theme = this._themeManager.theme(name),
                option = this.option(name);
            return isScalar ? void 0 !== option ? option : theme : $.extend(true, {}, theme, option)
        },
        _setupResizeHandler: function() {
            if (_parseScalar(this._getOption("redrawOnResize", true), true)) {
                this._addResizeHandler()
            } else {
                this._removeResizeHandler()
            }
        },
        _addResizeHandler: function() {
            var that = this;
            if (!that._resizeHandler) {
                that._resizeHandler = createResizeHandler(function() {
                    that._updateSize()
                });
                _windowResizeCallbacks.add(that._resizeHandler)
            }
        },
        _removeResizeHandler: function() {
            if (this._resizeHandler) {
                _windowResizeCallbacks.remove(this._resizeHandler);
                this._resizeHandler.dispose();
                this._resizeHandler = null
            }
        },
        beginUpdate: function() {
            var that = this;
            if (that._initialized && 0 === that._updateLockCount) {
                that._scheduleLoadingIndicatorHiding();
                that._changedOptions = createChangedOptions()
            }
            that.callBase.apply(that, arguments);
            return that
        },
        endUpdate: function() {
            var that = this,
                changedOptions = that._changedOptions;
            that.callBase.apply(that, arguments);
            if (changedOptions && 0 === that._updateLockCount) {
                that._changedOptions = null;
                if (changedOptions._num) {
                    that._renderer.lock();
                    that.beginUpdate();
                    that._handleChangedOptions(changedOptions);
                    that.endUpdate();
                    that._renderer.unlock()
                }
            }
            return that
        },
        _clean: $.noop,
        _render: $.noop,
        _optionChanged: function(args) {
            ++this._changedOptions._num;
            this._changedOptions[args.name] = args.value;
            this.callBase.apply(this, arguments)
        },
        _handleChangedOptions: function(options) {
            var that = this;
            $.each(options, function(name) {
                that._eventTrigger.update(name)
            });
            if ("redrawOnResize" in options) {
                that._setupResizeHandler()
            }
            if ("theme" in options || OPTION_RTL_ENABLED in options) {
                that._setThemeAndRtl()
            }
            if ("encodeHtml" in options) {
                that._handleThemeOptions()
            }
            if ("tooltip" in options) {
                that._setTooltipOptions()
            }
            if ("title" in options) {
                that._updateTitle()
            }
            if (OPTION_LOADING_INDICATOR in options) {
                that._updateLoadingIndicatorOptions()
            }
            if ("export" in options || "onExporting" in options || "onExported" in options || "onFileSaving" in options) {
                that._updateExportMenu()
            }
            if ("size" in options || "margin" in options) {
                that._updateSize()
            }
            if (options._has(that._invalidatingOptions || [])) {
                that._invalidate()
            }
        },
        _handleThemeOptions: function() {
            var that = this,
                options = {
                    rtl: that.option(OPTION_RTL_ENABLED),
                    encodeHtml: that.option("encodeHtml"),
                    animation: that._getAnimationOptions()
                };
            that._renderer.setOptions(options);
            that._setTooltipRendererOptions(options);
            that._setTooltipOptions();
            that._renderer.lock();
            that._updateTitle();
            that._updateExportMenuOptions();
            that._updateLoadingIndicatorOptions();
            that._scheduleLoadingIndicatorHiding();
            that._handleThemeOptionsCore();
            that._renderer.unlock()
        },
        _handleThemeOptionsCore: function() {
            this._initialized && this._refresh()
        },
        _visibilityChanged: function() {
            this.render()
        },
        _setThemeAndRtl: function() {
            this._themeManager.setTheme(this.option("theme"), this.option(OPTION_RTL_ENABLED))
        },
        _setTooltipRendererOptions: function(options) {
            this._tooltip.setRendererOptions(options)
        },
        _setTooltipOptions: function() {
            this._tooltip.update(this._getOption("tooltip"))
        },
        _updateTitle: function() {
            var isSizeChanged = this._title.update(this._getTitleOptions());
            if (isSizeChanged && this._initialized) {
                this._setContentSize()
            }
        },
        _updateExportMenu: function() {
            this._updateExportMenuOptions();
            this._initialized && this._setContentSize()
        },
        _getTitleOptions: function() {
            return $.extend(true, {}, this._themeManager.theme("title"), processTitleOptions(this.option("title")))
        },
        _hideTooltip: function() {
            this._tooltip.hide()
        },
        _normalizeHtml: function(html) {
            var re = /xmlns="[\s\S]*?"/gi,
                first = true;
            html = html.replace(re, function(match) {
                if (!first) {
                    return ""
                }
                first = false;
                return match
            });
            return html.replace(/xmlns:NS1="[\s\S]*?"/gi, "").replace(/NS1:xmlns:xlink="([\s\S]*?)"/gi, 'xmlns:xlink="$1"')
        },
        svg: function() {
            return this._normalizeHtml(this._renderer.svg())
        },
        isReady: getFalse,
        _dataIsReady: getTrue,
        _resetIsReady: function() {
            this.isReady = getFalse
        },
        _drawn: function() {
            var that = this;
            that.isReady = getFalse;
            if (that._dataIsReady()) {
                that._renderer.onEndAnimation(function() {
                    that.isReady = getTrue
                });
                that._fulfillLoadingIndicatorHiding()
            }
            that._eventTrigger("drawn", {})
        },
        _initExportMenu: function() {
            this._exportMenu = this._createExportMenu && this._createExportMenu()
        },
        _initHeaderBlock: function() {
            this._headerBlock = new headerBlockModule.HeaderBlock
        },
        _initLoadingIndicator: function() {
            var that = this;
            that._loadingIndicator = new loadingIndicatorModule.LoadingIndicator({
                eventTrigger: that._eventTrigger,
                renderer: that._renderer,
                notify: notify
            });
            that._scheduleLoadingIndicatorHiding();

            function notify(state) {
                that._skipLoadingIndicatorOptions = true;
                that.option(OPTION_LOADING_INDICATOR, {
                    show: state
                });
                that._skipLoadingIndicatorOptions = false;
                if (state) {
                    that._hideTooltip()
                }
            }
        },
        _disposeLoadingIndicator: function() {
            this._loadingIndicator.dispose();
            this._loadingIndicator = null
        },
        _disposeExportMenu: function() {
            if (this._exportMenu) {
                this._exportMenu.dispose();
                this._exportMenu = null
            }
        },
        _disposeHeaderBlock: function() {
            this._headerBlock.dispose();
            this._headerBlock = null
        },
        _updateLoadingIndicatorSize: function() {
            this._loadingIndicator.setSize(this._canvas)
        },
        _updateExportMenuOptions: function() {
            this._exportMenu && this._exportMenu.setOptions(this._getExportMenuOptions())
        },
        _updateLoadingIndicatorOptions: function() {
            if (!this._skipLoadingIndicatorOptions) {
                this._loadingIndicator.setOptions(this._getOption(OPTION_LOADING_INDICATOR))
            }
        },
        _scheduleLoadingIndicatorHiding: function() {
            this._loadingIndicator.scheduleHiding()
        },
        _fulfillLoadingIndicatorHiding: function() {
            this._loadingIndicator.fulfillHiding()
        },
        showLoadingIndicator: function() {
            this._loadingIndicator.show()
        },
        hideLoadingIndicator: function() {
            this._loadingIndicator.hide()
        },
        exportTo: function(fileName, format) {
            var creatorFunc, exportOptions = getExportOptions(this, fileName, format),
                exportMenu = this._exportMenu;
            if ("SVG" === exportOptions.format) {
                creatorFunc = clientExporter.svg.getBlob
            } else {
                if ("PDF" === exportOptions.format) {
                    creatorFunc = clientExporter.pdf.getBlob
                } else {
                    creatorFunc = clientExporter.image.getBlob
                }
            }
            exportMenu && exportMenu.hide();
            clientExporter.export(this.svg(), exportOptions, creatorFunc);
            exportMenu && exportMenu.show()
        },
        print: function() {
            var vizWindow = window.open();
            if (!vizWindow) {
                return
            }
            vizWindow.document.open();
            vizWindow.document.write(this.svg());
            vizWindow.document.close();
            vizWindow.print();
            vizWindow.close()
        },
        _createExportMenu: function() {
            var that = this;
            return new exportMenuModule.ExportMenu(that._renderer, function() {
                return that.svg()
            })
        },
        _getExportMenuOptions: function() {
            var that = this,
                userOptions = that._getOption("export") || {},
                options = getExportOptions(that, userOptions.fileName, userOptions.format, userOptions.color);
            return $.extend({}, userOptions, {
                exportOptions: options
            })
        }
    });

    function createChangedOptions() {
        return {
            _has: hasAnyOfFields,
            _num: 0
        }
    }

    function getExportOptions(widget, fileName, format, backgroundColor) {
        var validatedFormat = String(format).toUpperCase();
        if (-1 === $.inArray(validatedFormat, ["PNG", "PDF", "JPEG", "SVG", "GIF"])) {
            validatedFormat = "PNG"
        }
        return {
            format: validatedFormat,
            fileName: fileName || "file",
            proxyUrl: widget.option("export.proxyUrl"),
            width: widget._canvas.width,
            height: widget._canvas.height,
            exportingAction: widget._createActionByOption("onExporting"),
            exportedAction: widget._createActionByOption("onExported"),
            fileSavingAction: widget._createActionByOption("onFileSaving")
        }
    }
    var primaryAlignmentData = {};
    primaryAlignmentData.left = primaryAlignmentData.top = 0;
    primaryAlignmentData.right = primaryAlignmentData.bottom = 2;
    var secondaryAlignmentData = {};
    secondaryAlignmentData.left = secondaryAlignmentData.top = function(a, b, size) {
        return [a, a + size]
    };
    secondaryAlignmentData.right = secondaryAlignmentData.bottom = function(a, b, size) {
        return [b - size, b]
    };
    secondaryAlignmentData.center = function(a, b, size) {
        return [(a + b - size) / 2, (a + b + size) / 2]
    };

    function truncateLayoutArea(rect, layout) {
        var sideIndex = layout.side,
            sideOffset = primaryAlignmentData[layout.alignment[sideIndex]],
            side = layout.size[sideIndex];
        rect[sideIndex + sideOffset] += (sideOffset > 0 ? -1 : 1) * side
    }

    function buildupLayoutArea(rect, layout) {
        var sideIndex = layout.side,
            sideOffset = primaryAlignmentData[layout.alignment[sideIndex]],
            side = layout.size[sideIndex],
            otherSide = secondaryAlignmentData[layout.alignment[1 - sideIndex]](rect[1 - sideIndex], rect[3 - sideIndex], layout.size[1 - sideIndex]),
            result = [];
        result[sideIndex] = result[sideIndex + 2] = rect[sideIndex + sideOffset];
        result[sideIndex + sideOffset] = rect[sideIndex + sideOffset] -= (sideOffset > 0 ? -1 : 1) * side;
        result[1 - sideIndex] = otherSide[0];
        result[3 - sideIndex] = otherSide[1];
        return result
    }

    function hasAnyOfFields(fields) {
        var i, ii = fields.length;
        for (i = 0; !(fields[i] in this) && i < ii; ++i) {}
        return i < ii
    }

    function createEventTrigger(eventsMap, callbackGetter) {
        var triggers = {};
        $.each(eventsMap, function(name, info) {
            if (info.name) {
                createEvent(name)
            }
        });
        triggerEvent.update = function(name) {
            var eventInfo = eventsMap[name];
            if (eventInfo) {
                createEvent(eventInfo.newName || name)
            }
            return !!eventInfo
        };
        triggerEvent.dispose = function() {
            eventsMap = callbackGetter = triggers = null
        };
        return triggerEvent;

        function createEvent(name) {
            var eventInfo = eventsMap[name];
            triggers[eventInfo.name] = callbackGetter(name)
        }

        function triggerEvent(name, arg, complete) {
            triggers[name](arg);
            complete && complete()
        }
    }
});
