/** 
 * DevExtreme (viz/core/export.js)
 * Version: 16.1.4
 * Build date: Wed Jun 22 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var $ = require("jquery"),
        _extend = $.extend,
        _each = $.each,
        clientExporter = require("../../client_exporter"),
        messageLocalization = require("../../localization/message"),
        imageExporter = clientExporter.image,
        svgExporter = clientExporter.svg,
        pdfExporter = clientExporter.pdf,
        clickEvent = require("../../events/click"),
        pointerEvents = require("../../events/pointer"),
        pointerActions = [pointerEvents.down, pointerEvents.move].join(" "),
        BUTTON_SIZE = 35,
        ICON_COORDS = [
            [9, 12, 26, 12, 26, 14, 9, 14],
            [9, 17, 26, 17, 26, 19, 9, 19],
            [9, 22, 26, 22, 26, 24, 9, 24]
        ],
        LIST_WIDTH = 120,
        VERTICAL_TEXT_MARGIN = 8,
        HORIZONTAL_TEXT_MARGIN = 15,
        MARGIN = 10,
        SHADOW_OFFSET = 2,
        SHADOW_BLUR = 3,
        PRINTING_TEXT = messageLocalization.format("vizExport-printingButtonText"),
        ALLOWED_EXPORT_FORMATS = ["PNG", "PDF", "JPEG", "SVG", "GIF"],
        EXPORT_CSS_CLASS = "dx-export-menu",
        EXPORT_DATA_KEY = "export-element-type";

    function validateFormat(format) {
        var validatedFormat = String(format).toUpperCase();
        if (-1 !== $.inArray(validatedFormat, ALLOWED_EXPORT_FORMATS)) {
            return validatedFormat
        }
    }

    function getCreatorFunc(format) {
        if ("SVG" === format) {
            return svgExporter.getBlob
        } else {
            if ("PDF" === format) {
                return pdfExporter.getBlob
            } else {
                return imageExporter.getBlob
            }
        }
    }

    function print(data) {
        var vizWindow = window.open();
        if (!vizWindow) {
            return
        }
        vizWindow.document.open();
        vizWindow.document.write(data);
        vizWindow.document.close();
        vizWindow.print();
        vizWindow.close()
    }
    exports.exportFromMarkup = function(markup, options) {
        options.format = validateFormat(options.format) || "PNG";
        options.fileName = options.fileName || "file";
        clientExporter.export(markup, options, getCreatorFunc(options.format))
    };
    exports.ExportMenu = function(renderer, svgMethod) {
        var that = this;
        that._renderer = renderer;
        that._svgMethod = svgMethod;
        that._group = renderer.g().attr({
            "class": EXPORT_CSS_CLASS
        }).linkOn(renderer.root, "peripheral");
        that._buttonGroup = renderer.g().attr({
            "class": EXPORT_CSS_CLASS + "-button"
        }).append(that._group);
        that._listGroup = renderer.g().attr({
            "class": EXPORT_CSS_CLASS + "-list"
        }).append(that._group);
        that._subscribeEvents()
    };
    _extend(exports.ExportMenu.prototype, {
        getLayoutOptions: function() {
            var bbox = this._buttonGroup.getBBox();
            bbox.cutSide = "vertical";
            bbox.cutLayoutSide = "top";
            bbox.height += 2 * MARGIN;
            bbox.position = {
                vertical: "top",
                horizontal: "right"
            };
            bbox.verticalAlignment = "top";
            bbox.horizontalAlignment = "right";
            return bbox
        },
        probeDraw: $.noop,
        shift: function(_, y) {
            this._group.attr({
                translateY: this._group.attr("translateY") + y
            })
        },
        draw: function(_, height, canvas) {
            this._options.exportOptions.width = canvas.width;
            this._options.exportOptions.height = canvas.height;
            this._group.move(canvas.width - BUTTON_SIZE - SHADOW_OFFSET - SHADOW_BLUR, height - BUTTON_SIZE - MARGIN)
        },
        show: function() {
            this._group.linkAppend()
        },
        hide: function() {
            this._group.linkRemove()
        },
        setOptions: function(options) {
            this._options = options;
            options.formats = options.formats || ALLOWED_EXPORT_FORMATS;
            options.printingEnabled = void 0 === options.printingEnabled ? true : options.printingEnabled;
            if (options.enabled && (options.formats.length || options.printingEnabled)) {
                this.show();
                this._updateButton();
                this._updateList();
                this._hideList()
            } else {
                this.hide()
            }
        },
        dispose: function() {
            var that = this;
            that._unsubscribeEvents();
            that._group.linkRemove().linkOff();
            that._group.dispose();
            that._group = that._listGroup = that._buttonGroup = that._button = null;
            that._options = null
        },
        _hideList: function() {
            this._listGroup.attr({
                visibility: "hidden"
            });
            this._listShown = false;
            this._setButtonFill(this._options.menuButtonColor)
        },
        _showList: function() {
            this._listGroup.attr({
                visibility: "visible"
            });
            this._listShown = true
        },
        _setButtonFill: function(menuButtonColor) {
            this._button && this._button.attr({
                fill: menuButtonColor
            })
        },
        _subscribeEvents: function() {
            var that = this,
                isButtonSelected = false;
            that._renderer.root.on(clickEvent.name + ".export", function(e) {
                var exportOptions, elementType = e.target[EXPORT_DATA_KEY],
                    options = that._options;
                if (!elementType) {
                    that._hideList();
                    return
                }
                if ("button" === elementType) {
                    if (that._listShown) {
                        that._hideList();
                        isButtonSelected = false
                    } else {
                        that._showList();
                        that._setButtonFill(options.selectedColor);
                        isButtonSelected = true
                    }
                } else {
                    if ("printing" === elementType) {
                        that.hide();
                        print(that._svgMethod());
                        that.show();
                        that._hideList()
                    } else {
                        if ("exporting" === elementType) {
                            that.hide();
                            exportOptions = _extend({}, options.exportOptions, {
                                format: e.target["export-element-format"],
                                backgroundColor: options.backgroundColor
                            });
                            clientExporter.export(that._svgMethod(), exportOptions, getCreatorFunc(exportOptions.format));
                            that.show();
                            that._hideList()
                        }
                    }
                }
            });
            that._listGroup.on(pointerActions, function(e) {
                e.stopPropagation()
            });
            that._buttonGroup.on(pointerEvents.enter, function() {
                !isButtonSelected && that._setButtonFill(that._options.selectedColor)
            });
            that._buttonGroup.on(pointerEvents.leave, function() {
                !isButtonSelected && that._setButtonFill(that._options.menuButtonColor)
            })
        },
        _unsubscribeEvents: function() {
            this._renderer.root.off(".export");
            this._listGroup.off();
            this._buttonGroup.off()
        },
        _updateButton: function() {
            var icon, that = this,
                renderer = that._renderer,
                options = that._options,
                iconAttr = {
                    fill: options.font.color,
                    cursor: "pointer"
                },
                exportData = {
                    "export-element-type": "button"
                };
            if (!that._button) {
                that._button = renderer.rect(0, 0, BUTTON_SIZE, BUTTON_SIZE).append(that._buttonGroup);
                that._button.attr({
                    rx: 4,
                    ry: 4,
                    fill: options.menuButtonColor,
                    stroke: options.borderColor,
                    "stroke-width": 1,
                    cursor: "pointer"
                });
                that._button.data(exportData);
                icon = renderer.path(ICON_COORDS).append(that._buttonGroup);
                icon.attr(iconAttr);
                icon.data(exportData);
                that._buttonGroup.setTitle(messageLocalization.format("vizExport-titleMenuText"))
            }
        },
        _updateList: function() {
            var printingText, separator, rect, shadow, textBBox, yCoord, that = this,
                options = that._options,
                printingHeight = 2 * VERTICAL_TEXT_MARGIN + options.font.size,
                listPadding = BUTTON_SIZE + 3,
                formats = options.formats,
                listGroup = that._listGroup,
                renderer = that._renderer,
                textCss = {
                    "font-size": options.font.size,
                    "font-family": options.font.family,
                    fill: options.font.color,
                    "font-weight": options.font.weight,
                    cursor: "pointer"
                },
                texts = [],
                currentTextY = listPadding + VERTICAL_TEXT_MARGIN,
                xCoord = -LIST_WIDTH + BUTTON_SIZE;
            listGroup.clear();
            if (options.printingEnabled) {
                yCoord = listPadding + printingHeight / 2;
                printingText = renderer.text(PRINTING_TEXT, xCoord + HORIZONTAL_TEXT_MARGIN, 0).append(listGroup);
                textBBox = printingText.getBBox();
                printingText.css(textCss);
                printingText.attr({
                    "text-anchor": "start",
                    y: yCoord - textBBox.y / 2
                });
                printingText.data({
                    "export-element-type": "printing"
                });
                currentTextY += textBBox.height + VERTICAL_TEXT_MARGIN;
                if (formats.length) {
                    separator = renderer.path([xCoord, currentTextY, xCoord + LIST_WIDTH, currentTextY], "line");
                    separator.attr({
                        stroke: options.borderColor,
                        "stroke-width": 1,
                        cursor: "pointer",
                        sharp: "v"
                    });
                    separator.data({
                        "export-element-type": "list"
                    });
                    currentTextY += VERTICAL_TEXT_MARGIN
                }
            }
            _each(formats, function(_, format) {
                var text;
                format = validateFormat(format);
                if (format) {
                    text = renderer.text(messageLocalization.format("vizExport-exportButtonText", format), xCoord + HORIZONTAL_TEXT_MARGIN, 0).append(listGroup);
                    textBBox = text.getBBox();
                    text.css(textCss);
                    text.data({
                        "export-element-type": "exporting",
                        "export-element-format": format
                    });
                    text.attr({
                        "text-anchor": "start",
                        y: currentTextY - textBBox.y
                    });
                    texts.push(text);
                    currentTextY += textBBox.height + VERTICAL_TEXT_MARGIN
                }
            });
            shadow = renderer.shadowFilter("-50%", "-50%", "200%", "200%", SHADOW_OFFSET, 6, SHADOW_BLUR, options.shadowColor, .8);
            rect = renderer.rect(xCoord, listPadding, LIST_WIDTH, currentTextY - listPadding).append(listGroup);
            rect.attr({
                fill: options.menuButtonColor,
                stroke: options.borderColor,
                "stroke-width": 1,
                cursor: "pointer",
                rx: 4,
                ry: 4,
                filter: shadow.ref
            });
            rect.data({
                "export-element-type": "list"
            });
            separator && separator.append(listGroup);
            printingText && printingText.append(listGroup);
            _each(texts, function(_, text) {
                text.append(listGroup)
            })
        }
    })
});
