/** 
 * DevExtreme (viz/core/title.js)
 * Version: 16.1.4
 * Build date: Wed Jun 22 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var _Number = Number,
        _patchFontOptions = require("./utils").patchFontOptions,
        parseHorizontalAlignment = require("./utils").enumParser(["left", "center", "right"]),
        parseVerticalAlignment = require("./utils").enumParser(["top", "bottom"]),
        DEFAULT_MARGIN = 10,
        DEFAULT_GAP = 3;

    function hasText(text) {
        return !!(text && String(text).length > 0)
    }

    function processTitleLength(elem, text, width) {
        if (elem.attr({
                text: text
            }).applyEllipsis(width)) {
            elem.setTitle(text)
        }
    }

    function pickMarginValue(value) {
        return value >= 0 ? _Number(value) : DEFAULT_MARGIN
    }

    function validateMargin(margin) {
        var result;
        if (margin >= 0) {
            result = {
                left: _Number(margin),
                top: _Number(margin),
                right: _Number(margin),
                bottom: _Number(margin)
            }
        } else {
            margin = margin || {};
            result = {
                left: pickMarginValue(margin.left),
                top: pickMarginValue(margin.top),
                right: pickMarginValue(margin.right),
                bottom: pickMarginValue(margin.bottom)
            }
        }
        return result
    }

    function Title(params) {
        this._params = params;
        this._group = params.renderer.g().attr({
            "class": params.cssClass
        }).linkOn(params.renderer.root, {
            name: "title",
            after: "peripheral"
        });
        this._hasText = false
    }
    require("jquery").extend(Title.prototype, require("./layout_element").LayoutElement.prototype, {
        dispose: function() {
            var that = this;
            that._group.linkRemove();
            that._group.linkOff();
            if (that._titleElement) {
                that._clipRect.dispose();
                that._titleElement = that._subtitleElement = that._clipRect = null
            }
            that._params = that._group = that._options = null
        },
        _updateOptions: function(options) {
            this._options = options;
            this._options.horizontalAlignment = parseHorizontalAlignment(options.horizontalAlignment, "center");
            this._options.verticalAlignment = parseVerticalAlignment(options.verticalAlignment, "top");
            this._options.margin = validateMargin(options.margin)
        },
        _updateStructure: function() {
            var that = this,
                renderer = that._params.renderer,
                group = that._group,
                alignObj = {
                    align: that._options.horizontalAlignment
                };
            if (!that._titleElement) {
                that._titleElement = renderer.text().attr(alignObj).append(group);
                that._subtitleElement = renderer.text().attr(alignObj);
                that._clipRect = renderer.clipRect();
                group.attr({
                    clipId: that._clipRect.id
                })
            }
            group.linkAppend();
            hasText(that._options.subtitle.text) ? that._subtitleElement.append(group) : that._subtitleElement.remove()
        },
        _updateTexts: function() {
            var titleBox, y, that = this,
                options = that._options,
                subtitleOptions = options.subtitle,
                titleElement = that._titleElement,
                subtitleElement = that._subtitleElement,
                testText = "A";
            titleElement.attr({
                text: testText,
                y: 0
            }).css(_patchFontOptions(options.font));
            titleBox = titleElement.getBBox();
            that._titleTextY = titleBox.height + titleBox.y;
            titleElement.attr({
                text: options.text
            });
            titleBox = titleElement.getBBox();
            y = -titleBox.y;
            titleElement.attr({
                y: y
            });
            if (hasText(subtitleOptions.text)) {
                y += titleBox.height + titleBox.y;
                subtitleElement.attr({
                    text: subtitleOptions.text,
                    y: 0
                }).css(_patchFontOptions(subtitleOptions.font));
                y += -subtitleElement.getBBox().y - that._titleTextY + DEFAULT_GAP;
                subtitleElement.attr({
                    y: y
                })
            }
        },
        _updateBoundingRectAlignment: function() {
            var boundingRect = this._boundingRect,
                options = this._options;
            boundingRect.verticalAlignment = options.verticalAlignment;
            boundingRect.horizontalAlignment = options.horizontalAlignment;
            boundingRect.cutLayoutSide = options.verticalAlignment;
            boundingRect.cutSide = "vertical";
            boundingRect.position = {
                horizontal: options.horizontalAlignment,
                vertical: options.verticalAlignment
            }
        },
        update: function(options) {
            var that = this,
                _hasText = hasText(options.text),
                isLayoutChanged = _hasText || _hasText !== that._hasText;
            if (_hasText) {
                that._updateOptions(options);
                that._updateStructure();
                that._updateTexts();
                that._boundingRect = {};
                that._updateBoundingRect();
                that._updateBoundingRectAlignment()
            } else {
                that._group.linkRemove();
                that._boundingRect = null
            }
            that._hasText = _hasText;
            return isLayoutChanged
        },
        draw: function(width, height) {
            var layoutOptions, that = this;
            that._group.linkAppend();
            that._correctTitleLength(width);
            layoutOptions = that.getLayoutOptions();
            if (layoutOptions.width > width || layoutOptions.height > height) {
                that._params.incidentOccurred("W2103");
                that._group.linkRemove();
                that._boundingRect.width = that._boundingRect.height = 0
            }
            return that
        },
        probeDraw: function(width, height) {
            this.draw(width, height);
            return this
        },
        _correctTitleLength: function(width) {
            var that = this,
                options = that._options,
                margin = options.margin,
                maxWidth = width - margin.left - margin.right;
            processTitleLength(that._titleElement, options.text, maxWidth);
            that._subtitleElement && processTitleLength(that._subtitleElement, options.subtitle.text, maxWidth);
            that._updateBoundingRect()
        },
        getLayoutOptions: function() {
            return this._boundingRect || null
        },
        measure: function(size) {
            var result, boundingRect = this._boundingRect;
            if (boundingRect) {
                this.draw(size[0], size[1]);
                result = {
                    size: [boundingRect.width, boundingRect.height],
                    alignment: [boundingRect.horizontalAlignment, boundingRect.verticalAlignment],
                    side: 1
                }
            }
            return result || null
        },
        shift: function(x, y) {
            var that = this,
                box = that.getLayoutOptions();
            that._group.move(x - box.x, y - box.y);
            that._setClipRectSettings();
            return that
        },
        _setClipRectSettings: function() {
            var bbox = this.getLayoutOptions();
            this._clipRect.attr({
                x: bbox.x,
                y: bbox.y,
                width: bbox.width,
                height: bbox.height
            })
        },
        _updateBoundingRect: function() {
            var box, that = this,
                options = that._options,
                margin = options.margin,
                boundingRect = that._boundingRect;
            box = that._group.getBBox();
            box.height += margin.top + margin.bottom - that._titleTextY;
            box.width += margin.left + margin.right;
            box.x -= margin.left;
            box.y += that._titleTextY - margin.top;
            if (options.placeholderSize > 0) {
                box.height = options.placeholderSize
            }
            boundingRect.height = box.height;
            boundingRect.width = box.width;
            boundingRect.x = box.x;
            boundingRect.y = box.y
        }
    });
    exports.Title = Title
});
