/** 
 * DevExtreme (viz/tree_map/tree_map.base.js)
 * Version: 16.1.4
 * Build date: Wed Jun 22 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var BaseWidget = require("../core/base_widget"),
        common = require("./common"),
        Node = require("./node"),
        _getTilingAlgorithm = require("./tiling").getAlgorithm,
        _getColorizer = require("./colorizing").getColorizer,
        _patchFontOptions = require("../core/utils").patchFontOptions,
        _buildRectAppearance = common.buildRectAppearance,
        _buildTextAppearance = common.buildTextAppearance,
        _noop = require("jquery").noop,
        _extend = require("jquery").extend,
        _max = Math.max,
        directions = {
            lefttoprightbottom: [1, 1],
            leftbottomrighttop: [1, -1],
            righttopleftbottom: [-1, 1],
            rightbottomlefttop: [-1, -1]
        },
        _CHANGES = common.CHANGES;
    require("jquery").map(["NODES_CREATE", "TILE_SETTINGS", "GROUP_SETTINGS", "MAX_DEPTH", "NODES_RESET", "TILES", "LABELS", "TILING", "LABELS_LAYOUT"], common.addChangeCode);
    require("./tiling.squarified");
    require("./tiling").setDefaultAlgorithm("squarified");
    require("./colorizing.discrete");
    require("./colorizing").setDefaultColorizer("discrete");

    function pickPositiveInteger(val) {
        return val > 0 ? Math.round(val) : 0
    }
    var dxTreeMap = BaseWidget.inherit({
        _eventsMap: _extend({}, BaseWidget.prototype._eventsMap),
        _handlers: {
            beginBuildNodes: _noop,
            buildNode: _noop,
            endBuildNodes: _noop,
            setTrackerData: _noop,
            calculateState: function(options) {
                return _buildRectAppearance(options)
            }
        },
        _rootClass: "dxtm-tree-map",
        _rootClassPrefix: "dxtm",
        _getDefaultSize: function() {
            return {
                width: 400,
                height: 400
            }
        },
        _createThemeManager: function() {
            return new ThemeManager
        },
        _init: function() {
            var that = this;
            that._changes = that._changesLocker = 0;
            that._rectOffsets = {};
            that._context = {
                suspend: function() {
                    if (!that._applyingChanges) {
                        that._suspendChanges()
                    }
                },
                resume: function() {
                    if (!that._applyingChanges) {
                        that._resumeChanges()
                    }
                },
                change: function(code) {
                    that._change(code)
                },
                settings: [{}, {}],
                calculateState: that._handlers.calculateState,
                calculateLabelState: _buildTextAppearance
            };
            that._suspendChanges();
            that.callBase.apply(that, arguments);
            that._updateData();
            that._resumeChanges()
        },
        _initCore: function() {
            var that = this,
                renderer = that._renderer;
            that._createProxyType();
            that._tilesGroup = renderer.g().linkOn(renderer.root, "tiles").linkAppend();
            that._labelsGroup = renderer.g().attr({
                align: "left"
            }).linkOn(renderer.root, "labels").linkAppend()
        },
        _createProxyType: _noop,
        _disposeCore: function() {
            var that = this;
            that._filter && that._filter.dispose();
            that._labelsGroup.linkOff();
            that._tilesGroup.linkOff()
        },
        beginUpdate: function() {
            this.callBase.apply(this, arguments);
            this._suspendChanges()
        },
        endUpdate: function() {
            this._resumeChanges();
            this.callBase.apply(this, arguments)
        },
        _applySize: function(rect) {
            this._tilingRect = rect.slice();
            this._suspendChanges();
            this._change(_CHANGES.TILING);
            this._resumeChanges()
        },
        _handleChangedOptions: function(options) {
            var that = this;
            that._suspendChanges();
            that.callBase.apply(that, arguments);
            that._handleChangedOptionsCore(options);
            that._resumeChanges()
        },
        _handleChangedOptionsCore: function(options) {
            var that = this;
            if ("dataSource" in options) {
                that._updateData()
            }
            if ("valueField" in options || "childrenField" in options) {
                that._change(_CHANGES.NODES_CREATE)
            }
            if ("colorField" in options || "colorizer" in options) {
                that._change(_CHANGES.TILES)
            }
            if ("labelField" in options) {
                that._change(_CHANGES.LABELS)
            }
            if ("tile" in options) {
                that._change(_CHANGES.TILE_SETTINGS)
            }
            if ("group" in options) {
                that._change(_CHANGES.GROUP_SETTINGS)
            }
            if ("maxDepth" in options) {
                that._change(_CHANGES.MAX_DEPTH)
            }
            if ("layoutAlgorithm" in options || "layoutDirection" in options) {
                that._change(_CHANGES.TILING)
            }
        },
        _themeDependentChanges: _CHANGES.TILE_SETTINGS | _CHANGES.GROUP_SETTINGS | _CHANGES.MAX_DEPTH,
        _handleThemeOptionsCore: function() {
            this._suspendChanges();
            this._change(this._themeDependentChanges);
            this._resumeChanges()
        },
        _updateData: function() {
            this._suspendChanges();
            this._isDataExpected = true;
            this._updateDataSource()
        },
        _dataSourceChangedHandler: function() {
            var that = this;
            if (!that._isDataExpected) {
                that._suspendChanges()
            }
            that._isDataExpected = false;
            that._change(_CHANGES.NODES_CREATE);
            that._resumeChanges()
        },
        _change: function(code) {
            this._changes |= code
        },
        _suspendChanges: function() {
            ++this._changesLocker
        },
        _resumeChanges: function() {
            if (0 === --this._changesLocker && this._changes) {
                this._applyChanges()
            }
        },
        _applyChanges: function() {
            var that = this;
            that._applyingChanges = true;
            if (that._changes & _CHANGES.NODES_CREATE) {
                that._buildNodes()
            }
            that._applyLogicalChanges();
            if (that._changes & _CHANGES.NODES_RESET) {
                that._resetNodes()
            }
            if (that._changes & _CHANGES.TILES) {
                that._applyTilesAppearance()
            }
            if (that._changes & _CHANGES.LABELS) {
                that._applyLabelsAppearance()
            }
            if (that._changes & _CHANGES.TILING) {
                that._performTiling()
            }
            if (that._changes & _CHANGES.LABELS_LAYOUT) {
                that._performLabelsLayout()
            }
            that._changes = 0;
            that._applyingChanges = false;
            that._drawn();
            that._context.forceReset = false
        },
        _applyLogicalChanges: function() {
            var that = this,
                renderer = that._renderer;
            renderer.lock();
            if (that._changes & _CHANGES.TILE_SETTINGS) {
                that._changeTileSettings()
            }
            if (that._changes & _CHANGES.GROUP_SETTINGS) {
                that._changeGroupSettings()
            }
            if (that._changes & _CHANGES.MAX_DEPTH) {
                that._changeMaxDepth()
            }
            renderer.unlock()
        },
        _buildNodes: function() {
            var processedData, that = this,
                root = that._root = that._topNode = new Node;
            root._id = 0;
            root.parent = {};
            root.data = {};
            root.level = root.index = -1;
            root.ctx = that._context;
            root.label = null;
            that._nodes = [root];
            that._handlers.beginBuildNodes();
            processedData = that._processDataSourceItems(that._dataSource.items() || []);
            traverseDataItems(root, processedData.items, 0, {
                itemsField: !processedData.isPlain && that._getOption("childrenField", true) || "items",
                valueField: that._getOption("valueField", true) || "value",
                buildNode: that._handlers.buildNode,
                ctx: that._context,
                nodes: that._nodes
            });
            that._onNodesCreated();
            that._handlers.endBuildNodes();
            that._change(_CHANGES.NODES_RESET)
        },
        _onNodesCreated: common.empty,
        _processDataSourceItems: function(items) {
            return {
                items: items,
                isPlain: false
            }
        },
        _changeTileSettings: function() {
            var that = this,
                options = that._getOption("tile"),
                offsets = that._rectOffsets,
                borderWidth = pickPositiveInteger(options.border.width),
                edgeOffset = borderWidth / 2,
                innerOffset = 1 & borderWidth ? .5 : 0,
                labelOptions = options.label,
                settings = that._context.settings[0];
            that._change(_CHANGES.TILES | _CHANGES.LABELS);
            settings.state = that._handlers.calculateState(options);
            that._filter = that._filter || that._renderer.shadowFilter("-50%", "-50%", "200%", "200%");
            that._filter.attr(labelOptions.shadow);
            that._calculateLabelSettings(settings, labelOptions, that._filter.ref);
            if (offsets.tileEdge !== edgeOffset || offsets.tileInner !== innerOffset) {
                offsets.tileEdge = edgeOffset;
                offsets.tileInner = innerOffset;
                that._change(_CHANGES.TILING)
            }
        },
        _changeGroupSettings: function() {
            var that = this,
                options = that._getOption("group"),
                labelOptions = options.label,
                offsets = that._rectOffsets,
                borderWidth = pickPositiveInteger(options.border.width),
                edgeOffset = borderWidth / 2,
                innerOffset = 1 & borderWidth ? .5 : 0,
                headerHeight = 0,
                groupPadding = pickPositiveInteger(options.padding),
                settings = that._context.settings[1];
            that._change(_CHANGES.TILES | _CHANGES.LABELS);
            settings.state = that._handlers.calculateState(options);
            that._calculateLabelSettings(settings, labelOptions);
            if (options.headerHeight >= 0) {
                headerHeight = pickPositiveInteger(options.headerHeight)
            } else {
                headerHeight = settings.labelParams.height + 2 * pickPositiveInteger(labelOptions.paddingTopBottom)
            }
            if (that._headerHeight !== headerHeight) {
                that._headerHeight = headerHeight;
                that._change(_CHANGES.TILING)
            }
            if (that._groupPadding !== groupPadding) {
                that._groupPadding = groupPadding;
                that._change(_CHANGES.TILING)
            }
            if (offsets.headerEdge !== edgeOffset || offsets.headerInner !== innerOffset) {
                offsets.headerEdge = edgeOffset;
                offsets.headerInner = innerOffset;
                that._change(_CHANGES.TILING)
            }
        },
        _calculateLabelSettings: function(settings, options, filter) {
            var bbox = this._getTextBBox(options.font),
                paddingLeftRight = pickPositiveInteger(options.paddingLeftRight),
                paddingTopBottom = pickPositiveInteger(options.paddingTopBottom);
            settings.labelState = _buildTextAppearance(options, filter);
            settings.labelState.visible = !("visible" in options) || !!options.visible;
            settings.labelParams = {
                unitWidth: bbox.width,
                height: bbox.height,
                hOffset: paddingLeftRight,
                vOffset: -bbox.y + paddingTopBottom,
                rtlEnabled: this._getOption("rtlEnabled", true),
                paddingTopBottom: paddingTopBottom,
                paddingLeftRight: paddingLeftRight
            }
        },
        _changeMaxDepth: function() {
            var context = this._context,
                maxDepth = this._getOption("maxDepth", true),
                minLevel = this._topNode.level + 1,
                maxLevel = maxDepth >= 1 ? minLevel + Math.round(maxDepth) - 1 : 1 / 0;
            if (context.minLevel !== minLevel || context.maxLevel !== maxLevel) {
                context.minLevel = minLevel;
                context.maxLevel = maxLevel;
                this._change(_CHANGES.NODES_RESET)
            }
        },
        _resetNodes: function() {
            this._tilesGroup.clear();
            this._renderer.initHatching();
            this._context.forceReset = true;
            this._change(_CHANGES.TILES | _CHANGES.LABELS | _CHANGES.TILING)
        },
        _processNodes: function(context, process) {
            processNodes(context, this._topNode, process)
        },
        _applyTilesAppearance: function() {
            var that = this,
                colorizer = _getColorizer(that._getOption("colorizer"), that._themeManager, that._topNode);
            that._processNodes({
                renderer: that._renderer,
                group: that._tilesGroup,
                setTrackerData: that._handlers.setTrackerData,
                colorField: that._getOption("colorField", true) || "color",
                getColor: colorizer
            }, processTileAppearance)
        },
        _applyLabelsAppearance: function() {
            var that = this;
            that._labelsGroup.clear();
            that._processNodes({
                renderer: that._renderer,
                group: that._labelsGroup,
                setTrackerData: that._handlers.setTrackerData,
                labelField: that._getOption("labelField", true) || "name"
            }, processLabelAppearance);
            that._change(_CHANGES.LABELS_LAYOUT)
        },
        _performTiling: function() {
            var that = this,
                context = {
                    algorithm: _getTilingAlgorithm(that._getOption("layoutAlgorithm", true)),
                    directions: directions[String(that._getOption("layoutDirection", true)).toLowerCase()] || directions.lefttoprightbottom,
                    headerHeight: that._headerHeight,
                    groupPadding: that._groupPadding,
                    rectOffsets: that._rectOffsets
                };
            that._topNode.innerRect = that._tilingRect;
            calculateRects(context, that._topNode);
            that._processNodes(context, processTiling);
            that._change(_CHANGES.LABELS_LAYOUT);
            that._onTilingPerformed()
        },
        _onTilingPerformed: common.empty,
        _performLabelsLayout: function() {
            this._processNodes(null, processLabelsLayout)
        },
        _getTextBBox: function(fontOptions) {
            var bbox, renderer = this._renderer,
                text = this._textForCalculations || renderer.text("0", 0, 0);
            this._textForCalculations = text;
            text.css(_patchFontOptions(fontOptions)).append(renderer.root);
            bbox = text.getBBox();
            text.remove();
            return bbox
        }
    });

    function traverseDataItems(root, dataItems, level, params) {
        var node, i, dataItem, items, nodes = [],
            allNodes = params.nodes,
            ii = dataItems.length,
            totalValue = 0;
        for (i = 0; i < ii; ++i) {
            dataItem = dataItems[i];
            node = new Node;
            node._id = allNodes.length;
            node.ctx = params.ctx;
            node.parent = root;
            node.level = level;
            node.index = nodes.length;
            node.data = dataItem;
            params.buildNode(node);
            allNodes.push(node);
            nodes.push(node);
            items = dataItem[params.itemsField];
            if (items && items.length) {
                traverseDataItems(node, items, level + 1, params)
            }
            if (dataItem[params.valueField] > 0) {
                node.value = Number(dataItem[params.valueField])
            }
            totalValue += node.value
        }
        root.nodes = nodes;
        root.value = totalValue
    }

    function processNodes(context, root, process) {
        var node, i, nodes = root.nodes,
            ii = nodes.length;
        for (i = 0; i < ii; ++i) {
            node = nodes[i];
            process(context, node);
            if (node.isNode()) {
                processNodes(context, node, process)
            }
        }
    }

    function processTileAppearance(context, node) {
        node.color = node.data[context.colorField] || context.getColor(node) || node.parent.color;
        node.updateStyles();
        node.tile = !node.ctx.forceReset && node.tile || createTile[Number(node.isNode())](context, node);
        node.applyState()
    }
    var createTile = [createLeaf, createGroup];

    function createLeaf(context, node) {
        var tile = context.renderer.simpleRect().append(context.group);
        context.setTrackerData(node, tile);
        return tile
    }

    function createGroup(context, node) {
        var outer = context.renderer.simpleRect().append(context.group),
            inner = context.renderer.simpleRect().append(context.group);
        context.setTrackerData(node, inner);
        return {
            outer: outer,
            inner: inner
        }
    }

    function processLabelAppearance(context, node) {
        node.updateLabelStyle();
        if (node.labelState.visible) {
            createLabel(context, node, node.labelState, node.labelParams)
        }
    }

    function createLabel(context, currentNode, settings, params) {
        var textData = currentNode.data[context.labelField];
        currentNode.label = textData ? String(textData) : null;
        textData = currentNode.customLabel || currentNode.label;
        if (textData) {
            currentNode.text = context.renderer.text(textData).attr(settings.attr).css(settings.css).append(context.group);
            context.setTrackerData(currentNode, currentNode.text);
            currentNode.textWidth = params.unitWidth * textData.length
        }
    }
    var emptyRect = [0, 0, 0, 0];

    function calculateRects(context, root) {
        var i, nodes = root.nodes,
            items = [],
            rects = [],
            sum = 0,
            ii = items.length = rects.length = nodes.length;
        for (i = 0; i < ii; ++i) {
            sum += nodes[i].value;
            items[i] = {
                value: nodes[i].value,
                i: i
            }
        }
        if (sum > 0) {
            context.algorithm({
                items: items.slice(),
                sum: sum,
                rect: root.innerRect.slice(),
                isRotated: 1 & nodes[0].level,
                directions: context.directions
            })
        }
        for (i = 0; i < ii; ++i) {
            rects[i] = items[i].rect || emptyRect
        }
        root.rects = rects
    }

    function processTiling(context, node) {
        var headerHeight, rect = node.parent.rects[node.index],
            rectOffsets = context.rectOffsets;
        if (node.isNode()) {
            setRectAttrs(node.tile.outer, buildTileRect(rect, node.parent.innerRect, rectOffsets.headerEdge, rectOffsets.headerInner));
            rect = marginateRect(rect, context.groupPadding);
            headerHeight = Math.min(context.headerHeight, rect[3] - rect[1]);
            node.rect = [rect[0], rect[1], rect[2], rect[1] + headerHeight];
            setRectAttrs(node.tile.inner, marginateRect(node.rect, rectOffsets.headerEdge));
            rect[1] += headerHeight;
            node.innerRect = rect;
            calculateRects(context, node)
        } else {
            node.rect = rect;
            setRectAttrs(node.tile, buildTileRect(rect, node.parent.innerRect, rectOffsets.tileEdge, rectOffsets.tileInner))
        }
    }

    function marginateRect(rect, margin) {
        return [rect[0] + margin, rect[1] + margin, rect[2] - margin, rect[3] - margin]
    }

    function buildTileRect(rect, outer, edgeOffset, innerOffset) {
        return [rect[0] + (rect[0] === outer[0] ? edgeOffset : +innerOffset), rect[1] + (rect[1] === outer[1] ? edgeOffset : +innerOffset), rect[2] - (rect[2] === outer[2] ? edgeOffset : -innerOffset), rect[3] - (rect[3] === outer[3] ? edgeOffset : -innerOffset)]
    }

    function setRectAttrs(element, rect) {
        element.attr({
            x: rect[0],
            y: rect[1],
            width: _max(rect[2] - rect[0], 0),
            height: _max(rect[3] - rect[1], 0)
        })
    }

    function processLabelsLayout(context, node) {
        if (node.text && node.labelState.visible) {
            layoutTextNode(node, node.labelParams)
        }
    }

    function layoutTextNode(node, params) {
        var rect = node.rect,
            hOffset = params.hOffset;
        node.text.attr({
            x: params.rtlEnabled ? rect[2] - hOffset : rect[0] + hOffset,
            y: rect[1] + params.vOffset,
            visibility: node.textWidth + params.paddingLeftRight < rect[2] - rect[0] && params.height + params.paddingTopBottom < rect[3] - rect[1] ? "visible" : "hidden"
        })
    }
    var ThemeManager = require("../core/base_theme_manager").BaseThemeManager.inherit({
        _themeSection: "treeMap",
        _fontFields: ["tile.label.font", "group.label.font", "loadingIndicator.font", "title.font", "title.subtitle.font", "tooltip.font"]
    });
    require("../../core/component_registrator")("dxTreeMap", dxTreeMap);
    module.exports = dxTreeMap
});
